@@ -20,10 +20,11 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
 #include <string.h>
 
 #include "pango-attributes.h"
-#include "pango-utils.h"
+#include "pango-impl-utils.h"
 
 struct _PangoAttrList
 {
@@ -56,7 +57,7 @@ static PangoAttribute *pango_attr_size_new_internal (int                   size,
 
 /**
  * pango_attr_type_register:
- * @name: an identifier for the type. (Currently unused.)
+ * @name: an identifier for the type (currently unused.)
  * 
  * Allocate a new attribute type ID.
  * 
@@ -72,11 +73,12 @@ pango_attr_type_register (const gchar *name)
 
 /**
  * pango_attribute_copy:
- * @attr: a #PangoAttribute.
+ * @attr: a #PangoAttribute
  * 
  * Make a copy of an attribute.
  * 
- * Return value: a newly allocated #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attribute_copy (const PangoAttribute *attr)
@@ -139,8 +141,10 @@ pango_attr_string_copy (const PangoAttribute *attr)
 static void
 pango_attr_string_destroy (PangoAttribute *attr)
 {
-  g_free (((PangoAttrString *)attr)->value);
-  g_free (attr);
+  PangoAttrString *sattr = (PangoAttrString *)attr;
+
+  g_free (sattr->value);
+  g_slice_free (PangoAttrString, sattr);
 }
 
 static gboolean
@@ -154,7 +158,7 @@ static PangoAttribute *
 pango_attr_string_new (const PangoAttrClass *klass,
 		       const char           *str)
 {
-  PangoAttrString *result = g_new (PangoAttrString, 1);
+  PangoAttrString *result = g_slice_new (PangoAttrString);
 
   result->attr.klass = klass;
   result->value = g_strdup (str);
@@ -168,7 +172,8 @@ pango_attr_string_new (const PangoAttrClass *klass,
  * 
  * Create a new font family attribute.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_family_new (const char *family)
@@ -188,13 +193,15 @@ pango_attr_family_new (const char *family)
 static PangoAttribute *
 pango_attr_language_copy (const PangoAttribute *attr)
 {
-  return g_memdup (attr, sizeof (PangoAttrLanguage));
+  return pango_attr_language_new (((PangoAttrLanguage *)attr)->value);
 }
 
 static void
 pango_attr_language_destroy (PangoAttribute *attr)
 {
-  g_free (attr);
+  PangoAttrLanguage *lattr = (PangoAttrLanguage *)attr;
+
+  g_slice_free (PangoAttrLanguage, lattr);
 }
 
 static gboolean
@@ -210,7 +217,8 @@ pango_attr_language_equal (const PangoAttribute *attr1,
  * 
  * Create a new language tag attribute. 
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_language_new (PangoLanguage *language)
@@ -226,7 +234,7 @@ pango_attr_language_new (PangoLanguage *language)
 
   g_return_val_if_fail (language != NULL, NULL);
   
-  result = g_new (PangoAttrLanguage, 1);
+  result = g_slice_new (PangoAttrLanguage);
 
   result->attr.klass = &klass;
   result->value = language;
@@ -248,7 +256,9 @@ pango_attr_color_copy (const PangoAttribute *attr)
 static void
 pango_attr_color_destroy (PangoAttribute *attr)
 {
-  g_free (attr);
+  PangoAttrColor *cattr = (PangoAttrColor *)attr;
+
+  g_slice_free (PangoAttrColor, cattr);
 }
 
 static gboolean
@@ -269,7 +279,7 @@ pango_attr_color_new (const PangoAttrClass *klass,
 		      guint16               green,
 		      guint16               blue)
 {
-  PangoAttrColor *result = g_new (PangoAttrColor, 1);
+  PangoAttrColor *result = g_slice_new (PangoAttrColor);
   result->attr.klass = klass;
   result->color.red = red;
   result->color.green = green;
@@ -286,7 +296,8 @@ pango_attr_color_new (const PangoAttrClass *klass,
  * 
  * Create a new foreground color attribute.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_foreground_new (guint16 red,
@@ -311,7 +322,8 @@ pango_attr_foreground_new (guint16 red,
  * 
  * Create a new background color attribute.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_background_new (guint16 red,
@@ -339,7 +351,9 @@ pango_attr_int_copy (const PangoAttribute *attr)
 static void
 pango_attr_int_destroy (PangoAttribute *attr)
 {
-  g_free (attr);
+  PangoAttrInt *iattr = (PangoAttrInt *)attr;
+
+  g_slice_free (PangoAttrInt, iattr);
 }
 
 static gboolean
@@ -356,7 +370,7 @@ static PangoAttribute *
 pango_attr_int_new (const PangoAttrClass *klass,
 		    int                   value)
 {
-  PangoAttrInt *result = g_new (PangoAttrInt, 1);
+  PangoAttrInt *result = g_slice_new (PangoAttrInt);
   result->attr.klass = klass;
   result->value = value;
 
@@ -374,7 +388,9 @@ pango_attr_float_copy (const PangoAttribute *attr)
 static void
 pango_attr_float_destroy (PangoAttribute *attr)
 {
-  g_free (attr);
+  PangoAttrFloat *fattr = (PangoAttrFloat *)attr;
+
+  g_slice_free (PangoAttrFloat, fattr);
 }
 
 static gboolean
@@ -391,7 +407,7 @@ static PangoAttribute*
 pango_attr_float_new  (const PangoAttrClass *klass,
                        double                value)
 {
-  PangoAttrFloat *result = g_new (PangoAttrFloat, 1);
+  PangoAttrFloat *result = g_slice_new (PangoAttrFloat);
   result->attr.klass = klass;
   result->value = value;
 
@@ -412,7 +428,9 @@ pango_attr_size_copy (const PangoAttribute *attr)
 static void
 pango_attr_size_destroy (PangoAttribute *attr)
 {
-  g_free (attr);
+  PangoAttrSize *sattr = (PangoAttrSize *)attr;
+
+  g_slice_free (PangoAttrSize, sattr);
 }
 
 static gboolean
@@ -444,7 +462,7 @@ pango_attr_size_new_internal (int size,
     pango_attr_size_equal
   };
 
-  result = g_new (PangoAttrSize, 1);
+  result = g_slice_new (PangoAttrSize);
   result->attr.klass = absolute ? &absolute_klass : &klass;
   result->size = size;
   result->absolute = absolute;
@@ -458,7 +476,8 @@ pango_attr_size_new_internal (int size,
  * 
  * Create a new font-size attribute in fractional points.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_size_new (int size)
@@ -472,7 +491,8 @@ pango_attr_size_new (int size)
  * 
  * Create a new font-size attribute in device units.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  *
  * Since: 1.8
  **/
@@ -488,7 +508,8 @@ pango_attr_size_new_absolute (int size)
  * 
  * Create a new font slant style attribute.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_style_new (PangoStyle style)
@@ -509,7 +530,8 @@ pango_attr_style_new (PangoStyle style)
  * 
  * Create a new font weight attribute.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_weight_new (PangoWeight weight)
@@ -530,7 +552,8 @@ pango_attr_weight_new (PangoWeight weight)
  *
  * Create a new font variant attribute (normal or small caps)
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_variant_new (PangoVariant variant)
@@ -551,7 +574,8 @@ pango_attr_variant_new (PangoVariant variant)
  * 
  * Create a new font stretch attribute
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_stretch_new (PangoStretch  stretch)
@@ -580,7 +604,7 @@ pango_attr_font_desc_destroy (PangoAttribute *attr)
   PangoAttrFontDesc *desc_attr = (PangoAttrFontDesc *)attr;
 
   pango_font_description_free (desc_attr->desc);
-  g_free (attr);
+  g_slice_free (PangoAttrFontDesc, desc_attr);
 }
 
 static gboolean
@@ -595,13 +619,14 @@ pango_attr_font_desc_equal (const PangoAttribute *attr1,
 
 /**
  * pango_attr_font_desc_new:
- * @desc: 
+ * @desc: the font description
  * 
- * Create a new font description attribute. (This attribute
+ * Create a new font description attribute. This attribute
  * allows setting family, style, weight, variant, stretch,
- * and size simultaneously.)
+ * and size simultaneously.
  * 
- * Return value: 
+ * Return value:  the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_font_desc_new (const PangoFontDescription *desc)
@@ -613,7 +638,7 @@ pango_attr_font_desc_new (const PangoFontDescription *desc)
     pango_attr_font_desc_equal
   };
 
-  PangoAttrFontDesc *result = g_new (PangoAttrFontDesc, 1);
+  PangoAttrFontDesc *result = g_slice_new (PangoAttrFontDesc);
   result->attr.klass = &klass;
   result->desc = pango_font_description_copy (desc);
 
@@ -625,9 +650,10 @@ pango_attr_font_desc_new (const PangoFontDescription *desc)
  * pango_attr_underline_new:
  * @underline: the underline style.
  * 
- * Create a new underline-style object.
+ * Create a new underline-style attribute.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_underline_new (PangoUnderline underline)
@@ -652,7 +678,10 @@ pango_attr_underline_new (PangoUnderline underline)
  * modifies the color of underlines. If not set, underlines
  * will use the foreground color.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
+ *
+ * Since: 1.8
  **/
 PangoAttribute *
 pango_attr_underline_color_new (guint16 red,
@@ -673,9 +702,10 @@ pango_attr_underline_color_new (guint16 red,
  * pango_attr_strikethrough_new:
  * @strikethrough: %TRUE if the text should be struck-through.
  * 
- * Create a new font strike-through attribute.
+ * Create a new strike-through attribute.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_strikethrough_new (gboolean strikethrough)
@@ -700,7 +730,10 @@ pango_attr_strikethrough_new (gboolean strikethrough)
  * modifies the color of strikethrough lines. If not set, strikethrough
  * lines will use the foreground color.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
+ *
+ * Since: 1.8
  **/
 PangoAttribute *
 pango_attr_strikethrough_color_new (guint16 red,
@@ -724,7 +757,8 @@ pango_attr_strikethrough_color_new (guint16 red,
  * 
  * Create a new baseline displacement attribute.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_rise_new (int rise)
@@ -746,7 +780,8 @@ pango_attr_rise_new (int rise)
  * Create a new font size scale attribute. The base font for the
  * affected text will have its size multiplied by @scale_factor.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute*
 pango_attr_scale_new (double scale_factor)
@@ -773,7 +808,10 @@ pango_attr_scale_new (double scale_factor)
  * other fonts on the system that might contain the characters in the
  * text.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
+ *
+ * Since: 1.4
  **/
 PangoAttribute *
 pango_attr_fallback_new (gboolean enable_fallback)
@@ -795,7 +833,8 @@ pango_attr_fallback_new (gboolean enable_fallback)
  * 
  * Create a new letter-spacing attribute.
  * 
- * Return value: the new #PangoAttribute.
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  *
  * Since: 1.6
  **/
@@ -830,12 +869,12 @@ pango_attr_shape_copy (const PangoAttribute *attr)
 static void
 pango_attr_shape_destroy (PangoAttribute *attr)
 {
-  const PangoAttrShape *shape_attr = (PangoAttrShape *)attr;
+  PangoAttrShape *shape_attr = (PangoAttrShape *)attr;
   
   if (shape_attr->destroy_func)
     shape_attr->destroy_func (shape_attr->data);
   
-  g_free (attr);
+  g_slice_free (PangoAttrShape, shape_attr);
 }
 
 static gboolean
@@ -859,19 +898,20 @@ pango_attr_shape_equal (const PangoAttribute *attr1,
 /**
  * pango_attr_shape_new_with_data:
  * @ink_rect:     ink rectangle to assign to each character
- * @logical_rect: logical rectangle assign to each character
+ * @logical_rect: logical rectangle to assign to each character
  * @data:         user data pointer
  * @copy_func:    function to copy @data when the attribute
  *                is copied. If %NULL, @data is simply copied
  *                as a pointer.
  * @destroy_func: function to free @data when the attribute
- *                is freed, or %NULL.
+ *                is freed, or %NULL
  * 
  * Like pango_attr_shape_new(), but a user data pointer is also
- * provided; this pointer can be accessed when rendering later
+ * provided; this pointer can be accessed when later
  * rendering the glyph.
  * 
- * Return value: the newly created attribute
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  *
  * Since: 1.8
  **/
@@ -894,7 +934,7 @@ pango_attr_shape_new_with_data (const PangoRectangle  *ink_rect,
   g_return_val_if_fail (ink_rect != NULL, NULL);
   g_return_val_if_fail (logical_rect != NULL, NULL);
   
-  result = g_new (PangoAttrShape, 1);
+  result = g_slice_new (PangoAttrShape);
   result->attr.klass = &klass;
   result->ink_rect = *ink_rect;
   result->logical_rect = *logical_rect;
@@ -908,14 +948,15 @@ pango_attr_shape_new_with_data (const PangoRectangle  *ink_rect,
 /**
  * pango_attr_shape_new:
  * @ink_rect:     ink rectangle to assign to each character
- * @logical_rect: logical rectangle assign to each character
+ * @logical_rect: logical rectangle to assign to each character
  * 
  * Create a new shape attribute. A shape is used to impose a
  * particular ink and logical rect on the result of shaping a
  * particular glyph. This might be used, for instance, for
  * embedding a picture or a widget inside a PangoLayout.
  * 
- * Return value: the newly created attribute
+ * Return value: the newly allocated #PangoAttribute, which should be
+ *               freed with pango_attribute_destroy().
  **/
 PangoAttribute *
 pango_attr_shape_new (const PangoRectangle *ink_rect,
@@ -934,7 +975,7 @@ pango_attr_list_get_type (void)
   static GType our_type = 0;
   
   if (our_type == 0)
-    our_type = g_boxed_type_register_static ("PangoAttrList",
+    our_type = g_boxed_type_register_static (I_("PangoAttrList"),
 					     (GBoxedCopyFunc) pango_attr_list_copy,
 					     (GBoxedFreeFunc) pango_attr_list_unref);
 
@@ -944,14 +985,15 @@ pango_attr_list_get_type (void)
 /**
  * pango_attr_list_new:
  * 
- * Create a new empty attribute list with a reference count of 1.
+ * Create a new empty attribute list with a reference count of one.
  * 
- * Return value: the newly allocated #PangoAttrList.
+ * Return value: the newly allocated #PangoAttrList, which should
+ *               be freed with pango_attr_list_unref().
  **/
 PangoAttrList *
 pango_attr_list_new (void)
 {
-  PangoAttrList *list = g_new (PangoAttrList, 1);
+  PangoAttrList *list = g_slice_new (PangoAttrList);
 
   list->ref_count = 1;
   list->attributes = NULL;
@@ -966,7 +1008,9 @@ pango_attr_list_new (void)
  * 
  * Increase the reference count of the given attribute list by one.
  *
- * Return value: The attribute list passed in (since 1.10)
+ * Return value: The attribute list passed in
+ *
+ * Since: 1.10
  **/
 PangoAttrList *
 pango_attr_list_ref (PangoAttrList *list)
@@ -1008,7 +1052,7 @@ pango_attr_list_unref (PangoAttrList *list)
 
       g_slist_free (list->attributes);
 
-      g_free (list);
+      g_slice_free (PangoAttrList, list);
     }
 }
 
@@ -1016,9 +1060,11 @@ pango_attr_list_unref (PangoAttrList *list)
  * pango_attr_list_copy:
  * @list: a #PangoAttrList
  * 
- * Copy @list and return an identical, new list.
+ * Copy @list and return an identical new list.
  *
- * Return value: new attribute list
+ * Return value: the newly allocated #PangoAttrList, with a
+ *               reference count of one, which should
+ *               be freed with pango_attr_list_unref().
  **/
 PangoAttrList *
 pango_attr_list_copy (PangoAttrList *list)
@@ -1432,9 +1478,10 @@ pango_attr_list_splice (PangoAttrList *list,
  * @list: a #PangoAttrList
  * 
  * Create a iterator initialized to the beginning of the list.
+ * @list must not be modified until this iterator is freed.
  * 
- * Return value: a new #PangoIterator. @list must not be modified
- *               until this iterator is freed with pango_attr_iterator_destroy().
+ * Return value: the newly allocated #PangoAttrIterator, which should
+ *               be freed with pango_attr_iterator_destroy().
  **/
 PangoAttrIterator *
 pango_attr_list_get_iterator (PangoAttrList  *list)
@@ -1443,7 +1490,7 @@ pango_attr_list_get_iterator (PangoAttrList  *list)
 
   g_return_val_if_fail (list != NULL, NULL);
 
-  iterator = g_new (PangoAttrIterator, 1);
+  iterator = g_slice_new (PangoAttrIterator);
   iterator->next_attribute = list->attributes;
   iterator->attribute_stack = NULL;
 
@@ -1544,7 +1591,8 @@ pango_attr_iterator_next (PangoAttrIterator *iterator)
  * 
  * Copy a #PangoAttrIterator
  *
- * Return value: Copy of @iterator
+ * Return value: the newly allocated #PangoAttrIterator, which should
+ *               be freed with pango_attr_iterator_destroy().
  **/
 PangoAttrIterator *
 pango_attr_iterator_copy (PangoAttrIterator *iterator)
@@ -1553,7 +1601,7 @@ pango_attr_iterator_copy (PangoAttrIterator *iterator)
 
   g_return_val_if_fail (iterator != NULL, NULL);
 
-  copy = g_new (PangoAttrIterator, 1);
+  copy = g_slice_new (PangoAttrIterator);
 
   *copy = *iterator;
 
@@ -1574,7 +1622,7 @@ pango_attr_iterator_destroy (PangoAttrIterator *iterator)
   g_return_if_fail (iterator != NULL);
   
   g_list_free (iterator->attribute_stack);
-  g_free (iterator);
+  g_slice_free (PangoAttrIterator, iterator);
 }
 
 /**
@@ -1618,7 +1666,7 @@ pango_attr_iterator_get (PangoAttrIterator *iterator,
  * @iterator: a #PangoAttrIterator
  * @desc: a #PangoFontDescription to fill in with the current values.
  *        The family name in this structure will be set using
- *        pango_font_description_set_family_static using values from
+ *        pango_font_description_set_family_static() using values from
  *        an attribute in the #PangoAttrList associated with the iterator,
  *        so if you plan to keep it around, you must call:
  *        <literal>pango_font_description_set_family (desc, pango_font_description_get_family (desc))</literal>.
@@ -1778,8 +1826,10 @@ pango_attr_iterator_get_font (PangoAttrIterator     *iterator,
  * of @list for which @func returns %TRUE and inserts them into
  * a new list.
  * 
- * Return value: a newly allocated #PangoAttrList or %NULL if
+ * Return value: the new #PangoAttrList or %NULL if
  *  no attributes of the given types were found.
+ *
+ * Since: 1.2
  **/
 PangoAttrList *
 pango_attr_list_filter (PangoAttrList       *list,
@@ -1790,13 +1840,11 @@ pango_attr_list_filter (PangoAttrList       *list,
   PangoAttrList *new = NULL;
   GSList *tmp_list;
   GSList *prev;
-  GSList *new_attrs;
   
   g_return_val_if_fail (list != NULL, NULL);
 
   tmp_list = list->attributes;
   prev = NULL;
-  new_attrs = NULL;
   while (tmp_list)
     {
       GSList *next = tmp_list->next;
@@ -1847,6 +1895,8 @@ pango_attr_list_filter (PangoAttrList       *list,
  * Return value: a list of all attributes for the current range.
  *   To free this value, call pango_attribute_destroy() on
  *   each value and g_slist_free() on the list.
+ *
+ * Since: 1.2
  **/
 GSList *
 pango_attr_iterator_get_attrs (PangoAttrIterator *iterator)
