@@ -19,21 +19,18 @@
  * Boston, MA 02111-1307, USA.
  */
 
+#include <config.h>
 #include <errno.h>
 #include <math.h>
 #include <string.h>
 #include <stdlib.h>
 
-#include "config.h"
-
 #include "pango-font.h"
-#include "pango-utils.h"
+#include "pango-impl-utils.h"
 
 #include <glib/gstdio.h>
 
-#ifdef HAVE_FRIBIDI
-#include <fribidi/fribidi.h>
-#endif
+#include "mini-fribidi/fribidi.h"
 
 #ifndef HAVE_FLOCKFILE
 #  define flockfile(f) (void)1
@@ -60,6 +57,8 @@ struct PangoAlias
 
 static GHashTable *pango_aliases_ht = NULL;
 
+PangoWarningHistory _pango_warning_history = { FALSE, FALSE, FALSE };
+
 /**
  * pango_trim_string:
  * @str: a string
@@ -423,7 +422,7 @@ pango_scan_string (const char **pos, GString *out)
 gboolean
 pango_scan_int (const char **pos, int *out)
 {
-  int i = 0;
+  unsigned int i = 0;
   char buf[32];
   const char *p = *pos;
 
@@ -456,137 +455,71 @@ static GHashTable *config_hash = NULL;
 static void
 read_config_file (const char *filename, gboolean enoent_error)
 {
-  FILE *file;
-    
-  GString *line_buffer;
-  GString *tmp_buffer1;
-  GString *tmp_buffer2;
-  char *errstring = NULL;
-  const char *pos;
-  char *section = NULL;
-  int line = 0;
-
-  file = g_fopen (filename, "r");
-  if (!file)
+  GKeyFile *key_file = g_key_file_new();
+  GError *key_file_error = NULL;
+  gchar **groups;
+  gsize groups_count = 0;
+  guint group_index;
+  
+  if (!g_key_file_load_from_file(key_file,filename, 0, &key_file_error))
     {
-      if (errno != ENOENT || enoent_error)
-	g_printerr ("Pango:%s: Error opening config file: %s\n",
-		    filename, g_strerror (errno));
+      if (key_file_error)
+	{
+	  if (key_file_error->domain != G_FILE_ERROR || key_file_error->code != G_FILE_ERROR_NOENT || enoent_error)
+	    {
+	      g_warning ("error opening config file '%s': %s\n",
+			  filename, key_file_error->message);
+	    }
+	  g_error_free(key_file_error);
+	}
+      g_key_file_free(key_file);
       return;
     }
   
-  line_buffer = g_string_new (NULL);
-  tmp_buffer1 = g_string_new (NULL);
-  tmp_buffer2 = g_string_new (NULL);
-
-  while (pango_read_line (file, line_buffer))
+  groups = g_key_file_get_groups (key_file, &groups_count);
+  for (group_index = 0; group_index < groups_count; group_index++)
     {
-      line++;
-
-      pos = line_buffer->str;
-      if (!pango_skip_space (&pos))
-	continue;
-
-      if (*pos == '[')	/* Section */
+      gsize keys_count = 0;
+      const gchar *group = groups[group_index];
+      GError *keys_error = NULL;
+      gchar **keys;
+      
+      keys = g_key_file_get_keys(key_file, group, &keys_count, &keys_error);
+      
+      if (keys)
 	{
-	  pos++;
-	  if (!pango_skip_space (&pos) ||
-	      !pango_scan_word (&pos, tmp_buffer1) ||
-	      !pango_skip_space (&pos) ||
-	      *(pos++) != ']' ||
-	      pango_skip_space (&pos))
-	    {
-	      errstring = g_strdup ("Error parsing [SECTION] declaration");
-	      goto error;
-	    }
+	  guint key_index;
 
-	  section = g_strdup (tmp_buffer1->str);
-	}
-      else			/* Key */
-	{
-	  gboolean empty = FALSE;
-	  gboolean append = FALSE;
-	  char *v;
-
-	  if (!section)
+	  for (key_index = 0; key_index < keys_count; key_index++)
 	    {
-	      errstring = g_strdup ("A [SECTION] declaration must occur first");
-	      goto error;
-	    }
-
-	  if (!pango_scan_word (&pos, tmp_buffer1) ||
-	      !pango_skip_space (&pos))
-	    {
-	      errstring = g_strdup ("Line is not of the form KEY=VALUE or KEY+=VALUE");
-	      goto error;
-	    }
-	  if (*pos == '+')
-	    {
-	      append = TRUE;
-	      pos++;
-	    }
-
-	  if (*(pos++) != '=')
-	    {
-	      errstring = g_strdup ("Line is not of the form KEY=VALUE or KEY+=VALUE");
-	      goto error;
-	    }
-	    
-	  if (!pango_skip_space (&pos))
-	    {
-	      empty = TRUE;
-	    }
-	  else
-	    {
-	      if (!pango_scan_string (&pos, tmp_buffer2))
+	      const gchar *key = keys[key_index];
+	      GError *key_error = NULL;
+	      gchar *value =  g_key_file_get_value(key_file, group, key, &key_error);
+	      if (value != NULL)
 		{
-		  errstring = g_strdup ("Error parsing value string");
-		  goto error;
+		  g_hash_table_insert (config_hash,
+				       g_strdup_printf ("%s/%s", group, key),
+				       value);
 		}
-	      if (pango_skip_space (&pos))
+	      if (key_error)
 		{
-		  errstring = g_strdup ("Junk after value string");
-		  goto error;
+		  g_warning ("error getting key '%s/%s' in config file '%s'\n",
+			     group, key, filename);
+		  g_error_free(key_error);
 		}
 	    }
+	  g_strfreev(keys);
+	}
 
-	  g_string_prepend_c (tmp_buffer1, '/');
-	  g_string_prepend (tmp_buffer1, section);
-
-	  if (append)
-	    {
-	      /* Get any existing value */
-	      v = g_hash_table_lookup (config_hash, tmp_buffer1->str);
-	      if (v)
-		g_string_prepend (tmp_buffer2, v);
-	    }
-	      
-	  if (!empty)
-	    {
-	      g_hash_table_insert (config_hash,
-				   g_strdup (tmp_buffer1->str),
-				   g_strdup (tmp_buffer2->str));
-	    }
+      if (keys_error)
+	{
+	  g_warning ("error getting keys in group '%s' of config file '%s'\n",
+		     filename, group);
+	  g_error_free(keys_error);
 	}
     }
-      
-  if (ferror (file))
-    errstring = g_strdup (g_strerror(errno));
-  
- error:
-
-  if (errstring)
-    {
-      g_printerr ("Pango:%s:%d: %s\n", filename, line, errstring);
-      g_free (errstring);
-    }
-      
-  g_free (section);
-  g_string_free (line_buffer, TRUE);
-  g_string_free (tmp_buffer1, TRUE);
-  g_string_free (tmp_buffer2, TRUE);
-
-  fclose (file);
+  g_strfreev(groups);
+  g_key_file_free(key_file);
 }
 
 static void
@@ -752,7 +685,7 @@ pango_parse_style (const char *str,
       break;
     }
   if (warn)
-    g_warning ("Style must be normal, italic, or oblique");
+    g_warning ("style must be normal, italic, or oblique");
   
   return FALSE;
 }
@@ -760,7 +693,7 @@ pango_parse_style (const char *str,
 /**
  * pango_parse_variant:
  * @str: a string to parse.
- * @style: a #PangoVariant to store the result in.
+ * @variant: a #PangoVariant to store the result in.
  * @warn: if %TRUE, issue a g_warning() on bad input.
  * 
  * Parses a font variant. The allowed values are "normal"
@@ -799,14 +732,14 @@ pango_parse_variant (const char   *str,
     }
   
   if (warn)
-    g_warning ("Variant must be normal or small_caps");
+    g_warning ("variant must be normal or small_caps");
   return FALSE;
 }
 
 /**
  * pango_parse_weight:
  * @str: a string to parse.
- * @style: a #PangoWeight to store the result in.
+ * @weight: a #PangoWeight to store the result in.
  * @warn: if %TRUE, issue a g_warning() on bad input.
  * 
  * Parses a font weight. The allowed values are "heavy",
@@ -883,11 +816,11 @@ pango_parse_weight (const char  *str,
       {
 	char *end;
 
-	*weight = strtol (str, &end, 0);
+	*weight = strtol (str, &end, 10);
 	if (*end != '\0')
 	  {
 	    if (warn)
-	      g_warning ("Cannot parse numerical weight '%s'", str);
+	      g_warning ("failed parsing numerical weight '%s'", str);
 	    return FALSE;
 	  }
 	return TRUE;
@@ -895,14 +828,14 @@ pango_parse_weight (const char  *str,
     }
   
   if (warn)
-    g_warning ("Weight must be ultralight, light, normal, bold, ultrabold, heavy, or an integer");
+    g_warning ("weight must be ultralight, light, normal, bold, ultrabold, heavy, or an integer");
   return FALSE;
 }
 
 /**
  * pango_parse_stretch:
  * @str: a string to parse.
- * @style: a #PangoStretch to store the result in.
+ * @stretch: a #PangoStretch to store the result in.
  * @warn: if %TRUE, issue a g_warning() on bad input.
  * 
  * Parses a font stretch. The allowed values are 
@@ -992,7 +925,7 @@ pango_parse_stretch (const char   *str,
     }
 
   if (warn)
-    g_warning ("Stretch must be ultra_condensed, extra_condensed, condensed, semi_condensed, normal, semi_expanded, expanded, extra_expanded, or ultra_expanded");
+    g_warning ("stretch must be ultra_condensed, extra_condensed, condensed, semi_condensed, normal, semi_expanded, expanded, extra_expanded, or ultra_expanded");
   return FALSE;
 }
 
@@ -1002,7 +935,7 @@ pango_matrix_get_type (void)
   static GType our_type = 0;
   
   if (our_type == 0)
-    our_type = g_boxed_type_register_static ("PangoMatrix",
+    our_type = g_boxed_type_register_static (I_("PangoMatrix"),
 					     (GBoxedCopyFunc) pango_matrix_copy,
 					     (GBoxedFreeFunc) pango_matrix_free);
 
@@ -1011,37 +944,46 @@ pango_matrix_get_type (void)
 
 /**
  * pango_matrix_copy:
- * @matrix: a #PangoMatrix
+ * @matrix: a #PangoMatrix, can be %NULL
  * 
  * Copies a #PangoMatrix.
  * 
- * Return value: a copy of @matrix. The result must be freed with
- *  pango_matrix_free().
+ * Return value: the newly allocated #PangoMatrix, which should
+ *               be freed with pango_matrix_free(), or %NULL if
+ *               @matrix was %NULL.
  *
  * Since: 1.6
  **/
 PangoMatrix *
 pango_matrix_copy (const PangoMatrix *matrix)
 {
-  g_return_val_if_fail (matrix != NULL, NULL);
+  PangoMatrix *new_matrix;
 
-  return g_memdup (matrix, sizeof (PangoMatrix));
+  if (matrix)
+    {
+      new_matrix = g_slice_new (PangoMatrix);
+      *new_matrix = *matrix;
+    }
+  else
+    new_matrix = NULL;
+
+  return new_matrix;
 }
 
 /**
  * pango_matrix_free:
- * @matrix: a #PangoMatrix
+ * @matrix: a #PangoMatrix, or %NULL
  * 
  * Free a #PangoMatrix created with pango_matrix_copy().
+ * Does nothing if @matrix is %NULL.
  *
  * Since: 1.6
  **/
 void
 pango_matrix_free (PangoMatrix *matrix)
 {
-  g_return_if_fail (matrix != NULL);
-  
-  g_free (matrix);
+  if (matrix)
+    g_slice_free (PangoMatrix, matrix);
 }
 
 /**
@@ -1156,6 +1098,61 @@ pango_matrix_concat (PangoMatrix       *matrix,
   matrix->y0  = tmp.yx * new_matrix->y0 + tmp.yy * new_matrix->y0 + tmp.y0;
 }
 
+/**
+ * pango_matrix_get_font_scale_factor:
+ * @matrix: a #PangoMatrix, may be %NULL
+ * 
+ * Returns the scale factor of a matrix on the height of the font.
+ * That is, the scale factor in the direction perpendicular to the
+ * vector that the X coordinate is mapped to.
+ *
+ * Return value: the scale factor of @matrix on the height of the font,
+ * or 1.0 if @matrix is %NULL.
+ *
+ * Since: 1.12
+ **/
+double
+pango_matrix_get_font_scale_factor (const PangoMatrix *matrix)
+{
+/*
+ * Based on cairo-matrix.c:_cairo_matrix_compute_scale_factors()
+ *
+ * Copyright 2005, Keith Packard
+ */
+  double det;
+  
+  if (!matrix)
+    return 1.0;
+  
+  det = matrix->xx * matrix->yy - matrix->yx * matrix->xy;
+
+  if (det == 0)
+    {
+      return 0.0;
+    }
+  else
+    {
+      double x = matrix->xx;
+      double y = matrix->yx;
+      double major, minor;
+
+      major = sqrt (x*x + y*y);
+      
+      /*
+       * ignore mirroring
+       */
+      if (det < 0)
+	det = - det;
+      
+      if (major)
+	minor = det / major;
+      else 
+	minor = 0.0;
+
+      return minor;
+    }
+}
+
 static const char canon_map[256] = {
    0,   0,   0,   0,   0,   0,   0,   0,    0,   0,   0,   0,   0,   0,   0,   0, 
    0,   0,   0,   0,   0,   0,   0,   0,    0,   0,   0,   0,   0,   0,   0,   0, 
@@ -1206,6 +1203,7 @@ pango_language_copy (PangoLanguage *language)
 {
   return language; /* language tags are const */
 }
+
 static void
 pango_language_free (PangoLanguage *language)
 {
@@ -1218,7 +1216,7 @@ pango_language_get_type (void)
   static GType our_type = 0;
   
   if (our_type == 0)
-    our_type = g_boxed_type_register_static ("PangoLanguage",
+    our_type = g_boxed_type_register_static (I_("PangoLanguage"),
                                              (GBoxedCopyFunc)pango_language_copy,
                                              (GBoxedFreeFunc)pango_language_free);
   return our_type;
@@ -1278,7 +1276,7 @@ pango_language_from_string (const char *language)
  *            %NULL is allowed and matches nothing but '*'
  * @range_list: a list of language ranges, separated by ';' characters.
  *   each element must either be '*', or a RFC 3066 language range
- *   canonicalized as by pango_language_from_string().
+ *   canonicalized as by pango_language_from_string()
  * 
  * Checks if a language tag matches one of the elements in a list of
  * language ranges. A language tag is considered to match a range
@@ -1286,7 +1284,7 @@ pango_language_from_string (const char *language)
  * or the range is a prefix of the tag, and the character after the
  * tag is '-'.
  *
- * Returns: %TRUE if a match was found.
+ * Return value: %TRUE if a match was found.
  **/
 gboolean
 pango_language_matches (PangoLanguage *language,
@@ -1318,7 +1316,7 @@ pango_language_matches (PangoLanguage *language,
 }
 
 typedef struct {
-  const char *lang;
+  const char lang[4];
   const char *str;
 } LangInfo;
 
@@ -1402,35 +1400,110 @@ pango_language_get_sample_string (PangoLanguage *language)
   return result;
 }
 
-#ifdef HAVE_FRIBIDI
-
-gboolean
-pango_log2vis_get_embedding_levels (gunichar       *str,
-				    int             len,
-				    PangoDirection *pbase_dir,
-				    guint8         *embedding_level_list)
+/**
+ * pango_log2vis_get_embedding_levels:
+ * @text:      the text to itemize.
+ * @length:    the number of bytes (not characters) to process, or -1
+ *             if @text is nul-terminated and the legnth should be calculated.
+ * @pbase_dir: input base direction, and output resolved direction.
+ *
+ * This will return the bidirectional embedding levels of the input paragraph
+ * as defined by the Unicode Bidirectional Algorithm available at:
+ *
+ *   http://www.unicode.org/reports/tr9/
+ *
+ * If the input base direction is a weak direction, the direction of the
+ * characters in the text will determine the final resolved direction.
+ *
+ * Return value: a newly allocated array of embedding levels, one item per
+ *               character (not byte), that should be freed using g_free.
+ *
+ * Since: 1.4
+ */
+guint8 *
+pango_log2vis_get_embedding_levels (const gchar    *text,
+				    int             length,
+				    PangoDirection *pbase_dir)
 {
   FriBidiCharType fribidi_base_dir;
-  gboolean result;
+  guint8 *embedding_levels_list;
 
-  fribidi_base_dir = (*pbase_dir == PANGO_DIRECTION_LTR) ? FRIBIDI_TYPE_L : FRIBIDI_TYPE_R;
+  switch (*pbase_dir)
+    {
+    case PANGO_DIRECTION_LTR:
+    case PANGO_DIRECTION_TTB_RTL:
+      fribidi_base_dir = FRIBIDI_TYPE_L;
+      break;
+    case PANGO_DIRECTION_RTL:
+    case PANGO_DIRECTION_TTB_LTR:
+      fribidi_base_dir = FRIBIDI_TYPE_R;
+      break;
+    case PANGO_DIRECTION_WEAK_RTL:
+      fribidi_base_dir = FRIBIDI_TYPE_WR;
+      break;
+    /*
+    case PANGO_DIRECTION_WEAK_LTR:
+    case PANGO_DIRECTION_NEUTRAL:
+    */
+    default:
+      fribidi_base_dir = FRIBIDI_TYPE_WL;
+      break;
+    }
+
+#ifdef FRIBIDI_HAVE_UTF8
+  {
+    if (length < 0)
+      length = strlen (text);
+    embedding_levels_list = fribidi_log2vis_get_embedding_levels_new_utf8 (text, length, &fribidi_base_dir);
+  }
+#else
+  {
+    gunichar *text_ucs4;
+    int n_chars;
+    text_ucs4 = g_utf8_to_ucs4_fast (text, length, &n_chars);
+    embedding_levels_list = g_new (guint8, n_chars);
+    fribidi_log2vis_get_embedding_levels ((FriBidiChar*)text_ucs4, n_chars,
+					  &fribidi_base_dir,
+					  (FriBidiLevel*)embedding_levels_list);
+    g_free (text_ucs4);
+  }
+#endif
 
-  result = fribidi_log2vis_get_embedding_levels (str, len, &fribidi_base_dir,
-						 embedding_level_list);
-  
   *pbase_dir = (fribidi_base_dir == FRIBIDI_TYPE_L) ?  PANGO_DIRECTION_LTR : PANGO_DIRECTION_RTL;
 
-  return result;
+  return embedding_levels_list;
 }
 
-#endif /* HAVE_FRIBIDI */
+/**
+ * pango_unichar_direction:
+ * @ch: a Unicode character
+ *
+ * Determines the direction of a character; either
+ * %PANGO_DIRECTION_LTR, %PANGO_DIRECTION_RTL, or
+ * %PANGO_DIRECTION_NEUTRAL.
+ *
+ * Return value: the direction of the character, as used in the
+ * Unicode bidirectional algorithm.
+ */
+PangoDirection
+pango_unichar_direction (gunichar ch)
+{
+  FriBidiCharType fribidi_ch_type = fribidi_get_type (ch);
+
+  if (!FRIBIDI_IS_LETTER (fribidi_ch_type))
+    return PANGO_DIRECTION_NEUTRAL;
+  else if (FRIBIDI_IS_RTL (fribidi_ch_type))
+    return PANGO_DIRECTION_RTL;
+  else
+    return PANGO_DIRECTION_LTR;
+}
 
 /**
  * pango_get_mirror_char:
- * @ch: a unicode character
+ * @ch: a Unicode character
  * @mirrored_ch: location to store the mirrored character
  *
- * If @ch has the Unicode mirrored property and there is another unicode
+ * If @ch has the Unicode mirrored property and there is another Unicode
  * character that typically has a glyph that is the mirror image of @ch's
  * glyph, puts that character in the address pointed to by @mirrored_ch.
  *
@@ -1474,7 +1547,7 @@ alias_free (struct PangoAlias *alias)
 
   g_free (alias->families);
   
-  g_free (alias);
+  g_slice_free (struct PangoAlias, alias);
 }
 
 static void
@@ -1504,7 +1577,6 @@ read_alias_file (const char *filename)
 
   while (pango_read_line (file, line_buffer))
     {
-      gboolean empty = FALSE;
       gboolean append = FALSE;
       line++;
 
@@ -1531,22 +1603,15 @@ read_alias_file (const char *filename)
 	  goto error;
 	}
       
-      if (!pango_skip_space (&pos))
+      if (!pango_scan_string (&pos, tmp_buffer2))
 	{
-	  empty = TRUE;
+	  errstring = g_strdup ("Error parsing value string");
+	  goto error;
 	}
-      else
+      if (pango_skip_space (&pos))
 	{
-	  if (!pango_scan_string (&pos, tmp_buffer2))
-	    {
-	      errstring = g_strdup ("Error parsing value string");
-	      goto error;
-	    }
-	  if (pango_skip_space (&pos))
-	    {
-	      errstring = g_strdup ("Junk after value string");
-	      goto error;
-	    }
+	  errstring = g_strdup ("Junk after value string");
+	  goto error;
 	}
 
       alias_key.alias = g_ascii_strdown (tmp_buffer1->str, -1);
@@ -1556,7 +1621,7 @@ read_alias_file (const char *filename)
       
       if (!alias)
 	{
-	  alias = g_new0 (struct PangoAlias, 1);
+	  alias = g_slice_new0 (struct PangoAlias);
 	  alias->alias = alias_key.alias;
 	  
 	  g_hash_table_insert (pango_aliases_ht,
@@ -1598,7 +1663,7 @@ read_alias_file (const char *filename)
 
   if (errstring)
     {
-      g_printerr ("Pango:%s:%d: %s\n", filename, line, errstring);
+      g_warning ("error reading alias file: %s:%d: %s\n", filename, line, errstring);
       g_free (errstring);
     }
       
@@ -1644,13 +1709,11 @@ pango_load_aliases (void)
  * @fontname: an ascii string
  * @families: will be set to an array of font family names.
  *    this array is owned by pango and should not be freed.
+ * @n_families: will be set to the length of the @families array.
  *
- * Look up all user defined aliases for the alias #fontname.
- * The resulting font family names will be stored in #families,
- * and the number of families will be returned.
- * 
- * Return value: the number of font famillies stored in the #families argument.
- *   This value is owned by Pango and must not be freed.
+ * Look up all user defined aliases for the alias @fontname.
+ * The resulting font family names will be stored in @families,
+ * and the number of families in @n_families.
  **/
 void
 pango_lookup_aliases (const char   *fontname,
@@ -1689,6 +1752,8 @@ pango_lookup_aliases (const char   *fontname,
  *
  * Return value: The direction corresponding to the first strong character.
  * If no such character is found, then %PANGO_DIRECTION_NEUTRAL is returned.
+ *
+ * Since: 1.4
  */
 PangoDirection
 pango_find_base_dir (const gchar *text,
@@ -1717,12 +1782,14 @@ pango_find_base_dir (const gchar *text,
 
 /**
  * pango_is_zero_width:
- * @ch: a unicode character
+ * @ch: a Unicode character
  *
  * Checks @ch to see if it is a zero-width character that should not be
  * normally rendered on the screen.
  *
  * Return value: %TRUE if @ch is a zero-width character, %FALSE otherwise
+ *
+ * Since: 1.10
  */
 gboolean
 pango_is_zero_width (gunichar ch)
@@ -1748,3 +1815,37 @@ pango_is_zero_width (gunichar ch)
 		(ch >= 0x202A && ch <= 0x202E)
 	 )) || ch == 0xFEFF;
 }
+
+/**
+ * pango_quantize_line_geometry:
+ * @thickness: pointer to the thickness of a line, in Pango scaled units
+ * @position: corresponding position
+ *
+ * Quantizes the thickness and position of a line, typically an
+ * underline or strikethrough, to whole device pixels, that is integer
+ * multiples of %PANGO_SCALE. The purpose of this function is to avoid
+ * such lines looking blurry.
+ *
+ * Since: 1.12
+ */
+void
+pango_quantize_line_geometry (int *thickness,
+			      int *position)
+{
+  int thickness_pixels = (*thickness + PANGO_SCALE / 2) / PANGO_SCALE;
+  if (thickness_pixels == 0)
+    thickness_pixels = 1;
+  
+  if (thickness_pixels & 1)
+    {
+      int new_center = ((*position - *thickness / 2) & ~(PANGO_SCALE - 1)) + PANGO_SCALE / 2;
+      *position = new_center + (PANGO_SCALE * thickness_pixels) / 2;
+    }
+  else
+    {
+      int new_center = ((*position - *thickness / 2 + PANGO_SCALE / 2) & ~(PANGO_SCALE - 1));
+      *position = new_center + (PANGO_SCALE * thickness_pixels) / 2;
+    }
+
+  *thickness = thickness_pixels * PANGO_SCALE;
+}
